'use strict';

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

function _possibleConstructorReturn(self, call) { if (!self) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return call && (typeof call === "object" || typeof call === "function") ? call : self; }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function, not " + typeof superClass); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, enumerable: false, writable: true, configurable: true } }); if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

/**
 * Licensed Materials - Property of IBM
 * IBM Cognos Products: BI Cloud (C) Copyright IBM Corp. 2020
 * US Government Users Restricted Rights - Use, duplication or disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 */
define(['dashboard-analytics/util/DashboardFormatter', 'com/ibm/vida/utils/vida-utils'], function (DashboardFormatter, vidautils) {

	var msg_scenario = 'Scenario';
	var msg_base_scenario = 'Base';
	var msg_average = 'Average';
	var msg_sum = 'Sum';
	/** Caption of the virtual Row Column & DataItem. */
	var msg_row_column = 'Row';
	/** Used to create a caption for each virtual Row Tuple. */
	var msg_row_tuple = '{0}';

	/*
  * Added a dependency to the vida-utils StringEx utility class since it provides a useful format function.
  * Dashboard/LiveWidget may provide an alternative utility that may be a better alternative, if you want to avoid
  * the dependency on Vida's utilities.
  */
	var StringEx = vidautils.StringEx;

	var nextScenarioId = 0;

	/**
  * The what-if extensions takes a query result, which would normally be drawn in a visualization, as input data.
  * This means that the query result, to this extensions, is source data. An AI model can be trained upon it and
  * that model can be used to predict values. These predictions may be given to a visualization, which requires
  * a DataSet too. This is why this module encodes two types of data sets: a source table and visualization data.
  */

	/**
  * Check whether a value should be considered "missing"", e.g. it is not a valid number.
  * @param {unknown} _value The value to test.
  * @returns {boolean} Whether the value is considered missing.
  */
	function isMissingValue(_value) {
		return _value === undefined || _value === null || isNaN(_value);
	}

	/**
  * Translates a Dashboard aggregation type to a supported aggregation type for this module.
  * @param {string} _aggregation The aggregation as defined by Dashboard.
  * @returns {string} The aggregation supported by this module.
  */
	function toSupportedAggregation(_aggregation) {
		switch (_aggregation) {
			case 'avg':
			case 'sum':
				return _aggregation;
			default:
				return 'sum'; // Fall back to sum
		}
	}

	/**
  * SpecFormatter wraps Dashboard's formatSpec and complies wih VIPR's IFormatter interface
  */

	var SpecFormatter = function () {
		function SpecFormatter(_formatSpec) {
			_classCallCheck(this, SpecFormatter);

			Object.defineProperty(this, 'formatSpec', {
				enumerable: true,
				value: _formatSpec
			});
		}

		// IFormatter

		/**
   * Format a value. The type a formatter can format depends on the context.
   * @param {number} _value Value in a type the formatter can format.
   * @returns {string} The value, formatted as a String.
   */


		SpecFormatter.prototype.format = function format(_value) {
			return DashboardFormatter.format(_value, this.formatSpec);
		};

		return SpecFormatter;
	}();

	/**
  * A Column in a table.
  * It contains the (header) label of the column and its unique name.
  * There are two specializations of a Column:
  *  - MeasureColumn for a column containing numerical values.
  *  - CategoricalColumn for a column containing discrete (string) values, encoded by a Tuple.
  */


	var Column = function () {
		function Column(_type, _label, _id, _kind) {
			_classCallCheck(this, Column);

			Object.defineProperties(this, {
				type: {
					enumerable: true,
					value: _type
				},
				label: {
					enumerable: true,
					value: _label
				},
				id: {
					enumerable: true,
					value: _id
				},
				kind: {
					enumerable: true,
					value: _kind || null
				}
			});
		}

		Column.prototype.toString = function toString() {
			return this.label;
		};

		return Column;
	}();

	/**
  * Select columns of kind `'input'`.
  * @param {Column} _column The column to inspect.
  * @returns {boolean} Whether the column has the kind `'input'`.
  */


	function isInputColumn(_column) {
		return _column.kind === 'input';
	}

	/**
  * Select columns of kind `'target'`.
  * @param {Column} _column The column to inspect.
  * @returns {boolean} Whether the column has the kind `'target'`.
  */
	function isTargetColumn(_column) {
		return _column.kind === 'target';
	}

	var MeasureColumn = function (_Column) {
		_inherits(MeasureColumn, _Column);

		function MeasureColumn(_label, _id, _kind, _aggregation, _formatter) {
			_classCallCheck(this, MeasureColumn);

			var _this = _possibleConstructorReturn(this, _Column.call(this, 'measure', _label, _id, _kind));

			Object.defineProperties(_this, {
				aggregation: {
					enumerable: true,
					value: _aggregation
				},
				formatter: {
					enumerable: true,
					value: _formatter
				}
			});
			return _this;
		}

		return MeasureColumn;
	}(Column);

	/**
  * Vida works with the concept of decorations. This is required to implement selections/highlighting.
  */


	var Decorated = function () {
		function Decorated() {
			_classCallCheck(this, Decorated);
		}

		// IDecorated

		/**
   * Check for existence of a decoration.
   * @param {string} _name The name of the decoration.
   * @returns {Boolean} True if the decoration is present on the object, false otherwise.
   */
		Decorated.prototype.hasDecoration = function hasDecoration(_name) {
			return this.hasOwnProperty(_name);
		};

		/**
   * Get the value of a decoration.
   * @param {String} _name The name of the decoration.
   * @returns {any} The value of the decoration.
   */


		Decorated.prototype.getDecoration = function getDecoration(_name) {
			return this[_name];
		};

		return Decorated;
	}();

	var Tuple = function (_Decorated) {
		_inherits(Tuple, _Decorated);

		function Tuple(_label, _id) {
			_classCallCheck(this, Tuple);

			var _this2 = _possibleConstructorReturn(this, _Decorated.call(this));

			Object.defineProperties(_this2, {
				label: {
					enumerable: true,
					value: _label
				},
				id: {
					enumerable: true,
					value: _id
				}
			});
			return _this2;
		}

		/** Convenience method to make this object represent itself as String */


		Tuple.prototype.toString = function toString() {
			return this.label;
		};

		// ITuple

		/**
   * Get the index of the ItemClassSet this tuple belongs to.
   * Note: if the index is out of bounds (e.g. -1, or 0 with no sets), it is assumed the tuple doesn't belong to a set.
   * @returns {number} The index of the ItemClassSet.
   */


		Tuple.prototype.getItemClassSetIndex = function getItemClassSetIndex() {
			// No multi itemclass set support.
			return 0;
		};

		/**
   * Get the item at the specified item index. The number of items is defined by the number of ItemClasses of the set that the tuple belongs to.
   * @param {number} _index The item index.
   * @returns {com/ibm/vipr/data/IItem} The item.
   */


		Tuple.prototype.getItem = function getItem(_index) {
			// Shortcut: no stacking support, so simply return this.
			return _index === 0 ? this : null;
		};

		// IItem

		/**
   * Get the caption of the item class.
   * @param {string} _kind The FormatKind of caption to retrieve.
   * @returns {string} The caption of the item class.
   */


		Tuple.prototype.getCaption = function getCaption() /*_kind*/{
			// No support for different kinds
			return this.label;
		};

		/**
   * Get the unique name of the item class.
   * @returns {string} The unique name
   */


		Tuple.prototype.getUniqueName = function getUniqueName() {
			return this.id;
		};

		return Tuple;
	}(Decorated);

	var CategoricalColumn = function (_Column2) {
		_inherits(CategoricalColumn, _Column2);

		function CategoricalColumn(_label, _id, _kind, _tuples) {
			_classCallCheck(this, CategoricalColumn);

			var _this3 = _possibleConstructorReturn(this, _Column2.call(this, 'categorical', _label, _id, _kind));

			Object.defineProperty(_this3, 'tuples', {
				enumerable: true,
				value: _tuples
			});
			return _this3;
		}

		return CategoricalColumn;
	}(Column);

	/**
  * RowsColumn is a virtual column to encode a virtual Row Number categorcal.
  * Mapping this virtual column will lead to a categorical data item that has each row number as separate tuple.s
  */


	var RowsColumn = new CategoricalColumn(msg_row_column, '(row#)', 'rows', null);

	/**
  * SourceTable simply encodes the rows of a query result as tabular data.
  * It is used to create the machine learning model, and also as basis for DeltaTable to create a view that combines
  * the source data with predicted outputs.s
  */

	var SourceTable = function () {
		function SourceTable(_columns, _rowData) {
			_classCallCheck(this, SourceTable);

			Object.defineProperties(this, {
				columns: {
					enumerable: true,
					value: Object.freeze(_columns)
				},
				_rowData: {
					value: Object.freeze(_rowData)
				}
			});
		}

		/**
   * Get the column at the specified index.
   * Negative indices are taken from the end, e.g. index `-1` is the last column.
   * @param {number} _columnIndex The column index.
   * @returns {Column} The column definition.
   */
		SourceTable.prototype.getColumn = function getColumn(_columnIndex) {
			var index = Math.floor(_columnIndex) || 0;
			if (index < 0) {
				index += this.columnCount;
			}
			if (index < 0 || index >= this.columnCount) {
				throw new RangeError('_columnIndex ' + _columnIndex + ' is out of range');
			}
			return this.columns[index];
		};

		/**
   * Calculate the row offset in the 1 dimensional data array.
   * Accepts negative indices as "from end" and validates indices are in range.
   * @param {number} _rowIndex The row index.
   * @returns {number} The row offset in the `_rowData`.
   */


		SourceTable.prototype._rowOffset = function _rowOffset(_rowIndex) {
			var rowIndex = _rowIndex;
			if (rowIndex < 0) {
				rowIndex += this.rowCount;
			}
			if (rowIndex < 0 || rowIndex >= this.rowCount) {
				throw new RangeError('_rowIndex ' + _rowIndex + ' is out of range');
			}
			return rowIndex * this.columnCount;
		};

		/**
   * Get (part of) a row by index.
   * @param {number} _rowIndex The index of the row.
   * Negative values are interpreted as "from end", e.g. `-1` is the last row.
   * @param {number} _startColumn Optional: the first column to include. Defaults to `0`.
   * Negative values are interpreted as "from end", e.g. `-1` is the last column.
   * @param {number} _endColumn Optional: the end column to stop the row at. Defaults to `columnCount`.
   * Negative values are interpreted as "from end", e.g. `-1` is the last column.
   * @param {function} _transform Optional value transformation callback.
   * @returns {number[]} The values for the requested row (range).
   */


		SourceTable.prototype.getRow = function getRow(_rowIndex) {
			var _startColumn = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 0;

			var _endColumn = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : this.columnCount;

			var _transform = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : null;

			var rowOffset = this._rowOffset(_rowIndex);
			var start = _startColumn;
			if (start < 0) {
				start += this.columnCount;
			}
			if (start < 0 || start > this.columnCount) throw new RangeError('_startColumn ' + _startColumn + ' is out of range');
			var end = _endColumn;
			if (end < 0) {
				end += this.columnCount;
			}
			if (end < 0 || end > this.columnCount) throw new RangeError('_endColumn ' + _endColumn + ' is out of range');
			if (end < start) end = start;
			var result = new Array(end - start);
			for (var i = start; i < end; i++) {
				var value = this._rowData[rowOffset + i];
				result[i - start] = _transform ? _transform(value, this.columns[i], i) : value;
			}
			return result;
		};

		/**
   * Get a value by row and column index.
   * @param {number} _rowIndex
   * @param {number} _columnIndex
   * @param {function|null} _transform Optional transformation on the value.
   * @returns {number}
   */


		SourceTable.prototype.getValue = function getValue(_rowIndex, _columnIndex) {
			var _transform = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : null;

			var rowOffset = this._rowOffset(_rowIndex);
			var columnIndex = _columnIndex;
			if (columnIndex < 0) columnIndex += this.columnCount;
			if (columnIndex < 0 || columnIndex >= this.columnCount) throw new RangeError('_columnIndex ' + _columnIndex + ' is out of range');
			var value = this._rowData[rowOffset + columnIndex];
			return _transform ? _transform(value, this.columns[columnIndex], columnIndex) : value;
		};

		/**
   * Get the row data as 1 dimensional Array.
   * @returns {number[]} The row data as 1 dimensional array.
   */


		SourceTable.prototype.getRowData = function getRowData() {
			return this._rowData;
		};

		/**
   * Determine how many columns have the specified kind.
   * @param {string} _kind The kind of the column, e.g. 'input', 'target' or 'context'.
   * @returns {number} The amount of columns that match the specified kind.
   */


		SourceTable.prototype.countColumnKind = function countColumnKind(_kind) {
			return this.columns.reduce(function (_acc, _column) {
				return _acc + (_column.kind === _kind ? 1 : 0);
			}, 0);
		};

		/**
   * Calculate the domain of a Column
   * @param {number|Column} _columnIndex The index of the column, or the column itself.
   * @returns {object} An object with min, max, sum and avg properties.
   */


		SourceTable.prototype.calculateDomain = function calculateDomain(_columnIndex) {
			var columnCount = this.columnCount;
			var rowCount = this.rowCount;
			var min = 0;
			var max = 0;
			var sum = 0;
			var columnIndex = void 0;
			if (_columnIndex instanceof Column) {
				columnIndex = this.columns.indexOf(_columnIndex);
			} else {
				columnIndex = _columnIndex;
				if (columnIndex < 0) {
					columnIndex += columnCount;
				}
			}
			if (columnIndex < 0 || columnIndex >= columnCount) throw new RangeError('_columnIndex ' + _columnIndex + ' is out of range');
			for (var rowIndex = 0; rowIndex < rowCount; rowIndex++) {
				var value = this._rowData[rowIndex * columnCount + columnIndex];
				if (rowIndex === 0) {
					min = max = value;
				} else {
					if (min > value) {
						min = value;
					}
					if (max < value) {
						max = value;
					}
				}
				sum += value;
			}
			return {
				min: min,
				max: max,
				sum: sum,
				avg: rowCount > 0 ? sum / rowCount : 0
			};
		};

		_createClass(SourceTable, [{
			key: 'columnCount',
			get: function get() {
				return this.columns.length;
			}
		}, {
			key: 'rowCount',
			get: function get() {
				return this._rowData.length / this.columns.length;
			}
		}]);

		return SourceTable;
	}();

	/**
  * Scenario is a simple wrapper to contain the inputs, results and Tuple associated with a scenario.
  * It implements the Tuple interface so it can be used as a Tuple on a virtual categorical item describing the scenarios.
  */


	var Scenario = function (_Decorated2) {
		_inherits(Scenario, _Decorated2);

		function Scenario(_caption, _id, _inputs, _results) {
			_classCallCheck(this, Scenario);

			var _this4 = _possibleConstructorReturn(this, _Decorated2.call(this));

			Object.defineProperties(_this4, {
				caption: {
					enumerable: true,
					writable: true,
					value: _caption
				},
				id: {
					enumerable: true,
					writable: true,
					value: _id
				},
				inputs: {
					enumerable: true,
					writable: true,
					value: _inputs
				},
				results: {
					enumerable: true,
					writable: true,
					value: _results
				}
			});
			return _this4;
		}

		Scenario.prototype.update = function update(_caption, _inputs, _results) {
			this.caption = _caption;
			this.inputs = _inputs;
			this.results = _results;
			return this;
		};

		// ITuple

		/**
   * Get the index of the ItemClassSet this tuple belongs to.
   * Note: if the index is out of bounds (e.g. -1, or 0 with no sets), it is assumed the tuple doesn't belong to a set.
   * @returns {number} The index of the ItemClassSet.
   */


		Scenario.prototype.getItemClassSetIndex = function getItemClassSetIndex() {
			// No multi itemclass set support.
			return 0;
		};

		/**
   * Get the item at the specified item index. The number of items is defined by the number of ItemClasses of the set that the tuple belongs to.
   * @param {number} _index The item index.
   * @returns {com/ibm/vipr/data/IItem} The item.
   */


		Scenario.prototype.getItem = function getItem(_index) {
			// Shortcut: no stacking support, so simply return this.
			return _index === 0 ? this : null;
		};

		// IItem

		/**
   * Get the caption of the item class.
   * @param {string} _kind The FormatKind of caption to retrieve.
   * @returns {string} The caption of the item class.
   */


		Scenario.prototype.getCaption = function getCaption() /*_kind*/{
			// No support for different kinds
			return this.caption;
		};

		/**
   * Get the unique name of the item class.
   * @returns {string} The unique name
   */


		Scenario.prototype.getUniqueName = function getUniqueName() {
			return this.id;
		};

		return Scenario;
	}(Decorated);

	/** Special scenario instance used to represent the base scenario. It has no overridden data */


	var BASE_SCENARIO = new Scenario(msg_base_scenario, 'scenario.base', null, null);

	/**
  * DeltaTable creates a view of the Base (original data) combined with a new scenario.
  * To preserve memory, it doesn't actually create a single merged view, instead it tracks the overriding inputs and
  * results.
  */

	var DeltaTable = function () {
		function DeltaTable(_sourceColumns, _rowData) {
			_classCallCheck(this, DeltaTable);

			var columns = new Array(_sourceColumns.length + 1);
			/** scenarios is also used as tuples array for our virtual categorical scenarios array */
			var scenarios = [];
			scenarios.push(BASE_SCENARIO);
			// Insert a scenario categorical column as first column
			columns[0] = new CategoricalColumn(msg_scenario, 'scenario', 'scenario', scenarios);
			// Copy existing data items to index 1 and up.
			for (var i = 0; i < _sourceColumns.length; i++) {
				columns[i + 1] = _sourceColumns[i];
			}

			Object.defineProperties(this, {
				columns: {
					enumerable: true,
					value: Object.freeze(columns)
				},
				_rowData: {
					value: _rowData
				},
				_scenarios: {
					value: scenarios
				}
			});
		}

		DeltaTable.prototype.getScenarioAt = function getScenarioAt(_index) {
			return this._scenarios[_index < 0 ? this._scenarios.length + _index : _index] || null;
		};

		DeltaTable.prototype.addScenario = function addScenario(_caption, _inputs, _results) {
			if (!_inputs || !_results) {
				throw new TypeError('A scenario must have inputs and results');
			}
			var scenario = new Scenario(_caption, 'scenario.scen' + ++nextScenarioId, _inputs, _results);
			this._scenarios.push(scenario);
			return scenario;
		};

		DeltaTable.prototype.removeScenario = function removeScenario(_scenario) {
			var scenarioIndex = this._scenarios.indexOf(_scenario);
			// Do not allow removing the base scenario, e.g. exclude index 0.
			if (scenarioIndex > 0) {
				this.removeScenarioAt(scenarioIndex);
				return true;
			}
			return false;
		};

		DeltaTable.prototype.removeScenarioAt = function removeScenarioAt(_index) {
			var scenarioIndex = _index < 0 ? this._scenarios.length + _index : _index;
			if (scenarioIndex > 0 && scenarioIndex < this._scenarios.length) {
				return this._scenarios.splice(scenarioIndex, 1)[0];
			}
			return null;
		};

		DeltaTable.prototype.clearScenarios = function clearScenarios() {
			// The "base" Scenario and its Tuple must be preserved
			this._scenarios.length = 1;
		};

		/**
   * Iterate all rows with a call to callback per row.
   * @param {function} _callback Callback called with: `_row, _columns, _rowIndex`.
   * Warning: the row Array is re-used and updated to preserve memory. Values must be copied to preserve them.
   */


		DeltaTable.prototype.iterate = function iterate(_callback) {
			var sourceColumnCount = this.columns.length - 1;
			var row = new Array(sourceColumnCount + 1);
			var inputOffset = this.columns.findIndex(isInputColumn);
			var targetIndex = this.columns.findIndex(isTargetColumn);
			/** Warning: includes the virtual "base" scenario */
			var scenarioCount = this._scenarios.length;
			for (var baseRowIndex = 0, rowOffset = 0; rowOffset < this._rowData.length; baseRowIndex++, rowOffset += sourceColumnCount) {
				// Select the base scenario
				row[0] = 0;
				// copy the source row
				for (var i = 0; i < sourceColumnCount; i++) {
					row[i + 1] = this._rowData[rowOffset + i];
				}
				// Invoke the callback for the "base" row
				_callback(row, this.columns, baseRowIndex * scenarioCount, baseRowIndex);

				// start at index 1, scenario 0 is the virtual base scenario.
				for (var scenarioIndex = 1; scenarioIndex < scenarioCount; scenarioIndex++) {
					// The scenarios Array does not contain a "base" scenario, adjust the indexs
					var scenario = this._scenarios[scenarioIndex];
					// Select the scenario tuple
					row[0] = scenarioIndex;

					// Apply the Scenario's input values
					var inputs = scenario.inputs[baseRowIndex];
					for (var _i = 0; _i < scenario.inputs.length; _i++) {
						row[inputOffset + _i] = inputs[_i];
					}
					// Apply the Scenario's target value
					row[targetIndex] = scenario.results[baseRowIndex];
					// Invoke the callback for the Secnario row
					_callback(row, this.columns, baseRowIndex * scenarioCount + scenarioIndex, baseRowIndex);
				}
			}
		};

		_createClass(DeltaTable, [{
			key: 'scenarios',
			get: function get() {
				// Create a frozen copy of the internal Array, so it is clear this array should not be modified externally.
				return Object.freeze(this._scenarios.slice(0));
			}
		}, {
			key: 'columnCount',
			get: function get() {
				return this.columns.length;
			}

			/**
    * Get the row count of the source data (e.g. the base scenario).
    */

		}, {
			key: 'baseRowCount',
			get: function get() {
				return this._rowData.length / (this.columns.length - 1);
			}

			/**
    * Get the full row count including all scenarios.
    */

		}, {
			key: 'rowCount',
			get: function get() {
				return this.baseRowCount * this.scenarioCount;
			}

			/**
    * The amount of scenarios in the DeltaTable.
    * Always includes the base scenario, e.g. after clear this returns `1`.
    */

		}, {
			key: 'scenarioCount',
			get: function get() {
				return this._scenarios.length;
			}
		}]);

		return DeltaTable;
	}();

	/**
  * ColumnMapping allows mapping Column instances to slots. This cannot be visualized directly, a processing step
  * needs to happen to turn a Table + ColumnMapping into VIPRData + SlotMapping. This is performed in createVIPRData.
  */


	var ColumnMapping = function () {
		function ColumnMapping(_vizDef) {
			_classCallCheck(this, ColumnMapping);

			Object.defineProperties(this, {
				vizDef: {
					enumerable: true,
					value: _vizDef
				},
				_aggregate: {
					writable: true,
					value: false
				},
				_mapping: {
					value: new Array(_vizDef.slots.length).fill(null)
				}
			});
		}

		/**
   * Get whether aggregation is preferred or not.
   * Note that setting this to `false` for visualization that require aggregation has no effect.
   */


		ColumnMapping.prototype.add = function add(_slotName, _column) {
			var slot = this.vizDef.getSlot(_slotName);
			if (!slot) {
				throw new Error('There is no ' + _slotName + ' slot');
			}
			var slotIndex = this.vizDef.slots.indexOf(slot);

			var mapping = this._mapping[slotIndex];

			if (_column) {
				if (!(_column instanceof Column)) {
					throw new TypeError('_column must be a Column');
				}
				if (!mapping) {
					mapping = [];
					this._mapping[slotIndex] = mapping;
				}
				// Set/add the column
				if (mapping.indexOf(_column) < 0) {
					if (slot.multiDataItem) {
						mapping.push(_column);
					} else {
						mapping[0] = _column;
					}
				}
			} else if (mapping) {
				// Clear existing mapping
				this._mapping[slotIndex] = null;
			}
		};

		ColumnMapping.prototype.get = function get(_slotName) {
			var slotIndex = this.vizDef.slots.indexOf(this.vizDef.getSlot(_slotName));
			return slotIndex < 0 ? null : this._mapping[slotIndex];
		};

		_createClass(ColumnMapping, [{
			key: 'aggregate',
			get: function get() {
				return this._aggregate;
			},
			set: function set(_value) {
				this._aggregate = !!_value;
			}
		}]);

		return ColumnMapping;
	}();

	/**
  * A DataItem is the abstract base for categorical or continuous data items in VIPR.
  * A DataItem is much like a Column, but categorical data items in VIPR allow stacking.
  * (This is not supported in this implementation).
  */


	var DataItem = function (_Decorated3) {
		_inherits(DataItem, _Decorated3);

		function DataItem(_type, _caption, _uniqueName, _kind, _columnIndex) {
			_classCallCheck(this, DataItem);

			var _this5 = _possibleConstructorReturn(this, _Decorated3.call(this));

			Object.defineProperties(_this5, {
				type: {
					enumerable: true,
					value: _type
				},
				caption: {
					enumerable: true,
					value: _caption
				},
				uniqueName: {
					enumerable: true,
					value: _uniqueName
				},
				// Not used by ViPR
				kind: {
					enumerable: true,
					value: _kind
				},
				columnIndex: {
					enumerable: true,
					value: _columnIndex
				}
			});
			return _this5;
		}

		DataItem.prototype.toString = function toString() {
			return this.label;
		};

		// IDataItem

		/**
   * Report the type of the DataItem.
   * Note that based on the returned type, VIPR will "cast" the DataItem to either ICatDataItem or IContDataItem.
   * @returns {string} The DataItemType, e.g. `"cat"` or `"cont"`.
   */


		DataItem.prototype.getType = function getType() {
			return this.type;
		};

		// IItemClass

		/**
   * Get the caption of the item class.
   * @param {string} _kind The FormatKind of caption to retrieve.
   * @returns {string} The caption of the item class.
   */


		DataItem.prototype.getCaption = function getCaption() /*_kind*/{
			return this.caption;
		};

		/**
   * Get the subcaption of the item class.
   * @param {string} _kind The FormatKind of subcaption to retrieve.
   * @returns {string|null} The caption of the item class, or `null` if not applicable.
   */


		DataItem.prototype.getSubCaption = function getSubCaption() /*_kind*/{
			return null;
		};

		/**
   * Get the unique name of the item class.
   * @returns {string} The unique name.
   */


		DataItem.prototype.getUniqueName = function getUniqueName() {
			return this.uniqueName;
		};

		return DataItem;
	}(Decorated);

	var ContDataItem = function (_DataItem) {
		_inherits(ContDataItem, _DataItem);

		function ContDataItem(_caption, _uniqueName, _kind, _columnIndex, _formatter, _aggregation) {
			_classCallCheck(this, ContDataItem);

			var _this6 = _possibleConstructorReturn(this, _DataItem.call(this, 'cont', _caption, _uniqueName, _kind, _columnIndex));

			Object.defineProperties(_this6, {
				formatter: {
					enumerable: true,
					value: _formatter
				},
				aggregation: {
					enumerable: true,
					value: _aggregation
				},
				_hasAggregation: {
					writable: true,
					value: false
				}
			});
			return _this6;
		}

		ContDataItem.prototype.setAggregationState = function setAggregationState(_hasAggregation) {
			this._hasAggregation = !!_hasAggregation;
		};

		// IItemClass

		ContDataItem.prototype.getSubCaption = function getSubCaption() /*_kind*/{
			if (this._hasAggregation) {
				switch (this.aggregation) {
					case 'avg':
						return msg_average;
					case 'sum':
						return msg_sum;
				}
			}
			return _DataItem.prototype.getSubCaption.call(this);
		};

		// IContDataItem

		/**
   * Get the ItemClass to use as a header for this DataItem
   * @returns {com/ibm/vipr/data/IItemClass} The ItemClass defining this DataItem.
   */


		ContDataItem.prototype.getItemClass = function getItemClass() {
			// Shortcut: implemented by this object
			return this;
		};

		/**
   * Get a formatter for the specified FormatKind, to format generated values, for example values rendered on an axis label.
   * @param {String} The {@link com/ibm/vipr/data/FormatKind FormatKind} to get a Formatter for, e.g. FormatKind.Label (recommended for axis labels) or FormatKind.Data.
   * @returns {com/ibm/vipr/data/IFormatter}
   */


		ContDataItem.prototype.getFormatter = function getFormatter() /*_kind*/{
			// No FormatKind support, Dashboard only has a single formatter.
			return this.formatter;
		};

		/**
   * Get the domain of the data.
   * @param {com/ibm/vipr/api/ICapability} _aggregation The aggregation capability, or null if no aggregation is required.
   *      This aggregation should always match what was returned by getCapabilities (e.g. null if there was no aggregation capability, or same parameters if there was).
   * @returns {com/ibm/vipr/data/IContDomain|null} The min and max of the domain, or null to auto-determine from the data.
   */


		ContDataItem.prototype.getDomain = function getDomain() /*_aggregation, _channel*/{
			// Returning null means VIPR will calculate the domain. That's OK since all rows are available.
			return null;
		};

		return ContDataItem;
	}(DataItem);

	var CatDataItem = function (_DataItem2) {
		_inherits(CatDataItem, _DataItem2);

		function CatDataItem(_caption, _uniqueName, _kind, _columnIndex, _tuples) {
			_classCallCheck(this, CatDataItem);

			var _this7 = _possibleConstructorReturn(this, _DataItem2.call(this, 'cat', _caption, _uniqueName, _kind, _columnIndex));

			Object.defineProperty(_this7, 'tuples', {
				enumerable: true,
				value: _tuples
			});
			return _this7;
		}

		// ICatDataItem

		/**
   * The number of ItemClassSets on the DataItem.
   * @returns {number}
   */


		CatDataItem.prototype.getItemClassSetCount = function getItemClassSetCount() {
			// Not supported; item class sets allow assymetrical query results.
			return 1;
		};

		/**
   * Get the set at the specified index.
   * @param {number} _index The zero-based index of the ItemClassSet to retrieve.
   * @returns {com/ibm/vipr/data/IItemClassSet} The ItemClassSet.
   */


		CatDataItem.prototype.getItemClassSet = function getItemClassSet(_index) {
			// Not supported, simply return this
			return _index === 0 ? this : null;
		};

		/**
   * The total number of tuples on the DataItem. This should indicate the total number of tuples on an DataItem, **not** the
   * number of Tuples in the current data window. When iterating Tuples, VIPR will respect the
   * {@link com/ibm/vipr/data/ICatDataItem#getDataWindowStart getDataWindowStart()} and {@link com/ibm/vipr/data/ICatDataItem#getDataWindowEnd getDataWindowEnd()} values.
   * @returns {Number}
   */


		CatDataItem.prototype.getTupleCount = function getTupleCount() {
			return this.tuples.length;
		};

		/**
   * The data window start of the DataItem.
   * The start index is the index of the first tuple that can be retrieved on this DataItem.
   * @returns {number}
   */


		CatDataItem.prototype.getDataWindowStart = function getDataWindowStart() {
			// No windowing support
			return 0;
		};

		/**
   * The data window end of the DataItem. Defined as a tuple range from start (inclusive) to end (exclusive).
   * The end index is the first out of bounds index (so end - start is the amount of tuples in the data window).
   * @returns {number}
   */


		CatDataItem.prototype.getDataWindowEnd = function getDataWindowEnd() {
			// No windowing support
			return this.tuples.length;
		};

		/**
   * Get a tuple at the specified tuple index.
   * VIPR will not call getTuple with an index that is outside of the data window.
   * @param {number} _indexOnDataItem The index of the tuple, relative to the absolute start of the ICatDataItem.
   * @returns {Tuple|undefined} The tuple at the specified index.
   */


		CatDataItem.prototype.getTuple = function getTuple(_indexOnDataItem) {
			return this.tuples[_indexOnDataItem];
		};

		// IItemClassSet

		/**
   * The number of item classes of this set. The item classes of a set basically define the
   * dimensionality of a set of tuples.
   * @returns {number}
   */


		CatDataItem.prototype.getItemClassCount = function getItemClassCount() {
			// No stacking support for this implementation, so return 1
			return 1;
		};

		/**
   * Get the item class at the specified index.
   * @param {number} _index The index of the item class.
   * @returns {com/ibm/vipr/data/IItemClass} The item class.
   */


		CatDataItem.prototype.getItemClass = function getItemClass(_index) {
			// Anotther shortcut; no stacking is supported so return this object itself.
			return _index === 0 ? this : null;
		};

		return CatDataItem;
	}(DataItem);

	var Value = function (_Decorated4) {
		_inherits(Value, _Decorated4);

		function Value(_dataItem) {
			_classCallCheck(this, Value);

			var _this8 = _possibleConstructorReturn(this, _Decorated4.call(this));

			Object.defineProperties(_this8, {
				_di: {
					value: _dataItem
				},
				// Value (summed)
				_v: {
					writable: true,
					value: 0
				},
				// Count; number of values added
				_c: {
					writable: true,
					value: 0
				}
			});
			return _this8;
		}

		/**
   * Add a value to the Value object.
   * @param {number|null} _value
   */


		Value.prototype.add = function add(_value) {
			if (typeof _value === 'number' && !Number.isNaN(_value)) {
				this._v += _value;
				this._c++;
			}
		};

		/**
   * The type of value.
   * @returns {string} A vipr ValueType.
   */


		Value.prototype.getValueType = function getValueType() {
			return this._c < 1 ? 'missing' : 'numeric';
		};

		/**
   * Get the value as a formatted string. This method should always return a correct value to
   * display as a String, for example an empty string for Missing.
   * @param {String} _kind ({@link com/ibm/vipr/data/FormatKind FormatKind}) The kind of format.
   * @returns {String} The formatted value.
   */


		Value.prototype.getCaption = function getCaption() /*_kind*/{
			return this._di.formatter.format(this.getValue());
		};

		/**
   * Gets the underlying value in its native type.
   * @member {Number|String|Date}
   */


		Value.prototype.getValue = function getValue() {
			if (this._c < 1) return null;
			switch (this._di.aggregation) {
				case 'sum':
					return this._v;
				case 'avg':
					return this._v / this._c;
			}
		};

		return Value;
	}(Decorated);

	var DataPoint = function (_Decorated5) {
		_inherits(DataPoint, _Decorated5);

		function DataPoint(_dataItems, _rowData, _offset) {
			_classCallCheck(this, DataPoint);

			var _this9 = _possibleConstructorReturn(this, _Decorated5.call(this));

			Object.defineProperties(_this9, {
				dataItems: {
					enumerable: true,
					value: _dataItems
				},
				_rowData: {
					value: _rowData
				},
				_offset: {
					value: _offset
				}
			});
			return _this9;
		}

		/**
   * Validate the data item index, since going out of bounds may actually put us in a different data point.
   * @param {number} _dataItemIndex
   * @returns {number} A valid index.
   */


		DataPoint.prototype._validateIndex = function _validateIndex(_dataItemIndex) {
			if (_dataItemIndex < 0 || _dataItemIndex >= this.dataItems.length) {
				throw new RangeError('Data item index out of range');
			}
			return _dataItemIndex;
		};

		/**
   * Get the index of the tuple for the specified DataItem.
   * Will only be called for indices that are reported to have a categorical data item.
   * @param {number} _dataItemIndex The index of DataItem to get the tuple index for.
   * @returns {number} The index of the Tuple on the associated CatDataItem.
   */


		DataPoint.prototype.getTupleIndex = function getTupleIndex(_dataItemIndex) {
			return this._rowData[this._offset + this._validateIndex(_dataItemIndex)];
		};

		/**
   * Get the value for the specified DataItem.
   * Will only be called for indices that are reported to have a continuous data item.
   * @param {Number} _dataItemIndex The index of DataItem to get the value for.
   * @returns {com/ibm/vipr/data/IValue} The value at the specified index (for the associated {@link com/ibm/vipr/data/IContDataItem IContDataItem}),
   * or null if there is no value.
   */


		DataPoint.prototype.getValue = function getValue(_dataItemIndex) {
			return this._rowData[this._offset + this._validateIndex(_dataItemIndex)];
		};

		return DataPoint;
	}(Decorated);

	/**
  * Generic Array iterator to comply with VIPR's iterator based API for data points.
  */


	var DataPointIterator = function () {
		function DataPointIterator(_array) {
			_classCallCheck(this, DataPointIterator);

			// This is basically a "self destructing" iterator, as soon as it reaches the end it drops the reference to the array
			Object.defineProperties(this, {
				_i: {
					writable: true,
					value: -1
				},
				_a: {
					writable: true,
					value: _array
				}
			});
		}

		DataPointIterator.prototype.nextValue = function nextValue() {
			var result = null;
			if (this._a) {
				var i = ++this._i;

				if (i < this._a.length) result = this._a[i];else this._a = null;
			}
			return result;
		};

		return DataPointIterator;
	}();

	/**
  * A DataSet is VIPR's name for a query result. The view is a hybrid between tabular and multi-dimensional data.
  * "Columns" are called DataItems, which can be categorical (discrete) or continuous (scalar).
  * This class also implements IDataSetMapping, since it's easy to combine the mapping with the data.
  */


	var VIPRDataSet = function (_Decorated6) {
		_inherits(VIPRDataSet, _Decorated6);

		function VIPRDataSet(_dataItems, _dataPoints, _slots, _mapping) {
			_classCallCheck(this, VIPRDataSet);

			var _this10 = _possibleConstructorReturn(this, _Decorated6.call(this));

			Object.defineProperties(_this10, {
				dataItems: {
					enumerable: true,
					value: _dataItems
				},
				dataPoints: {
					enumerable: true,
					value: _dataPoints
				},
				slots: {
					enumerable: true,
					value: _slots
				},
				mapping: {
					enumerable: true,
					value: _mapping
				}
			});
			return _this10;
		}

		// IDataSet

		/**
   * Get the number of DataItems in the data set.
   * @returns {Number} The number of DataItems in the data.
   */


		VIPRDataSet.prototype.getDataItemCount = function getDataItemCount() {
			return this.dataItems.length;
		};

		/**
   * Get the DataItem at the specified edge index.
   * @param {Number} _index The index of the data item to get.
   * @returns {com/ibm/vipr/data/IDataItem} The data item at the specified index.
   */


		VIPRDataSet.prototype.getDataItem = function getDataItem(_index) {
			return this.dataItems[_index];
		};

		/**
   * Get the data point iterator.
   * @returns {com/ibm/vipr/data/IDataPointIterator} The data point iterator.
   */


		VIPRDataSet.prototype.getDataPointIterator = function getDataPointIterator() {
			return new DataPointIterator(this.dataPoints);
		};

		// IDataSetMapping

		/**
   * Get the indices of data items mapped to the specified slot.
   * Return an empty array to not map any items (`null` is also allowed).
   * @param {com/ibm/vipr/vizdef/SlotRef} _slotRef The slot reference definition.
   * @returns {number[]|null} The array of DataItem indices to map, or `null` (equivalent of empty array)
   */


		VIPRDataSet.prototype.getDataItemsFor = function getDataItemsFor(_slotRef) {
			var slotIndex = this.slots.indexOf(_slotRef);
			return slotIndex < 0 ? null : this.mapping[slotIndex];
		};

		return VIPRDataSet;
	}(Decorated);

	/**
  * VIPRData is the root class to provide data to VIPR. It's basically a placeholder for DataSets, which is what
  * VIPR calls a "query result". This object also implements ISlotMapping, so it can be passed as slot mapping too.
  */


	var VIPRData = function () {
		function VIPRData(_dataSets) {
			_classCallCheck(this, VIPRData);

			Object.defineProperties(this, {
				dataSets: {
					enumerable: true,
					value: _dataSets
				}
			});
		}

		// IVIPRData

		/**
   * Get the data set at the specified index.
   * @param {number} _index The index of the data set.
   * @returns {VIPRDataSet} The data set.
   */


		VIPRData.prototype.getDataSetAt = function getDataSetAt(_index) {
			return this.dataSets[_index];
		};

		/**
   * Get the version of the decorations. If the decorations on VIPR data are updated, change
   * (e.g. increment) the returned version number to notify VIPR of this. Doing so does not
   * require a setData call, it will be picked up the next time decorations are requested.
   * @returns {number} A number representing the version of decorations.
   */


		VIPRData.prototype.getDecorationsVersion = function getDecorationsVersion() {
			return 0; // Not yet supported.
		};

		// ISlotMapping

		/**
   * Get the slot mapping for the data set at the specified index.
   * @param {number} _index The index of the data set.
   * @returns {VIPRDataSet} The data set mapping.
   */


		VIPRData.prototype.getDataSetMapping = function getDataSetMapping(_index) {
			// Just like VIPRData implements ISlotMapping, VIPRDataSet implements IDataSetMapping.
			return this.dataSets[_index];
		};

		return VIPRData;
	}();

	/**
  * Helper method to find the index of a DataItem that maps to the specified `_columnIndex`.
  * @param {DataItem[]} _dataItems The array of data items to search.
  * @param {number} _columnIndex The index of the desired referenced column
  * @returns {number} The resolved index, or `-1` if no data item matched.
  */


	function findDataItemIndex(_dataItems, _columnIndex) {
		for (var i = 0; i < _dataItems.length; i++) {
			if (_dataItems[i].columnIndex === _columnIndex) {
				return i;
			}
		}
		return -1;
	}

	/**
  * Create a DataItem from a Column
  * @param {Column} _column The column this DataItem is based on.
  * @param {number} _columnIndex The index of the column this data item is based on.
  * @returns {DataItem} The data item.
  */
	function createDataItem(_column, _columnIndex) {
		if (_column.type === 'measure') {
			return new ContDataItem(_column.label, _column.id, _column.kind, _columnIndex, _column.formatter, toSupportedAggregation(_column.aggregation));
		} else /*if (_column.type === 'categorical')*/{
				return new CatDataItem(_column.label, _column.id, _column.kind, _columnIndex,
				// Copy tuples if they are not frozen (should only be for the Scenario)
				Object.isFrozen(_column.tuples) ? _column.tuples : Object.freeze(_column.tuples.slice(0)));
			}
	}

	/**
  * Create a virtual Rows DataItem for the specified rowCount
  * @param {number} _rowCount The row count to generate tuples for.
  * @returns {CatDataItem} The virtual data item.
  */
	function createRowsDataItem(_rowCount) {
		var tuples = new Array(_rowCount);
		for (var i = 0; i < _rowCount; i++) {
			tuples[i] = new Tuple(StringEx.format(msg_row_tuple, i), 'row#' + i);
		}
		return new CatDataItem(RowsColumn.label, RowsColumn.id, RowsColumn.kind, -1, Object.freeze(tuples));
	}

	/**
  * Query a visualization data set, aggregating over unmapped categoricals as needed
  * @param {DataSetDef} _dataSetDef The VIPR DataSet definition.
  * @param {DeltaTable} _table The table to iterate for data.
  * @param {ColumnMapping} _columnMapping The column
  */
	function queryDataSet(_dataSetDef, _table, _columnMapping) {
		/** The mapping array contains the actual slot mapping. */
		var mapping = new Array(_dataSetDef.slots.length).fill(null);
		/** The data items as presented to the visualization */
		var dataItems = [];
		// Track if there is a virtual Rows DataItem.
		var rowsDataItem = null;
		_dataSetDef.slots.forEach(function (_slotRef, _index) {
			var columnMapping = _columnMapping.get(_slotRef.name);
			if (columnMapping) {
				/*
     * The column mapping contains Column instances.
     * VIPR wants DataItem indices in the mapping.
     * Translate the mapping to data item indices, creating a DataItem for a column if it doesn't exist.
     */
				mapping[_index] = columnMapping.map(function (_column) {
					var columnIndex = void 0;
					if (_column === RowsColumn) {
						// Use index -1 to encode the virtual rows column
						columnIndex = -1;
					} else {
						columnIndex = _table.columns.indexOf(_column);
						// In this case the -1 value is invalid, it should not default to the virtual rows column.
						if (columnIndex < 0) {
							// Program error, this should not happen.
							throw new Error('Mapped column is not part of the provided table');
						}
					}
					// Check if the column is already mapped to a data item
					var dataItemIndex = findDataItemIndex(dataItems, columnIndex);
					// If not, map a new data item
					if (dataItemIndex < 0) {
						var dataItem = void 0;
						if (columnIndex >= 0) {
							dataItem = createDataItem(_column, columnIndex);
						} else if (columnIndex === -1) {
							dataItem = createRowsDataItem(_table.baseRowCount);
							rowsDataItem = dataItem;
						}
						// Add the data item, and update the resolved index for the mapping.
						dataItemIndex = dataItems.push(dataItem) - 1;
					}
					return dataItemIndex;
				});
			}
		});

		// This step is optional: add unmapped continuous items to the data points too. This makes it easier to show them (aggregated) in a tooltip.
		for (var columnIndex = 0; columnIndex < _table.columns.length; columnIndex++) {
			var column = _table.columns[columnIndex];
			if (column.type === 'measure' && findDataItemIndex(dataItems, columnIndex) < 0) {
				// Add the unmapped cont data item
				dataItems.push(createDataItem(column, columnIndex));
			}
		}

		Object.freeze(dataItems);

		var volume = 1;
		var catItems = [];
		var ordinalOffsets = [];
		for (var dataItemIndex = 0; dataItemIndex < dataItems.length; dataItemIndex++) {
			var dataItem = dataItems[dataItemIndex];
			if (dataItem.type === 'cat' && dataItem.tuples.length !== 1) {
				catItems.push(dataItem);
				ordinalOffsets.push(volume);
				volume *= dataItem.tuples.length;
			}
		}

		/**
   * Many visualizations require aggregation, encoded by the uniqueDataPoints flag on the DataSetDef.
   * If visualizations support un-aggregated data points, check the preference set on the ColumnMapping.
   *
   * In this DeltaTable, there is one combination in which we know aggregation can never occur:
   * if both the virtual row and scenario columns are mapped, each DataPoint will be unique.
   *
   * This effectively means this flag can become `false` even though the visualization requires aggregation.
   */
		var shouldAggregate = (_dataSetDef.uniqueDataPoints || _columnMapping.aggregate) && ( // Must aggregate (by definition)?
		!rowsDataItem || findDataItemIndex(dataItems, 0) < 0); // Is aggregation even "possible"?
		/**
   * Maps ordinal to DataPoint. This is a map since the space may be sparse, e.g. have many missing rows.
   * If this cache is `null`, it means logic determined no aggregation was needed.
   */
		var dataPointCache = shouldAggregate ? new Map() : null;
		// If there is no aggregation, the amount of data points is known. In that case, initialize with the correct capacity.
		var dataPoints = dataPointCache ? new Array(_table.baseRowCount) : [];
		dataPoints.length = 0;
		// Allocate data. If there is aa rowsDataItem the amount of rows is known exactly.
		// If not, start with a capacity for 16 rows and let the array grow as needed.
		var rowData = new Array(dataItems.length * (dataPointCache ? _table.baseRowCount : 16));
		rowData.length = 0;

		/** Flag that tracks wether at least one value is aggregated. (If not, the aggregation captions can be avoided). */
		var hasAggregatedData = false;

		_table.iterate(function (_row, _columns, _rowIndex, _baseRowIndex) {
			// Start by calculating the ordinal for this row
			var ordinal = void 0;
			if (dataPointCache) {
				// Calculate the ordinal based on the mapped categorical items.
				ordinal = 0;
				for (var i = 0; i < catItems.length; ++i) {
					var tupleIndex = catItems[i] === rowsDataItem ? _baseRowIndex : _row[catItems[i].columnIndex];
					ordinal += tupleIndex * ordinalOffsets[i];
				}
			} else {
				// If there is no need to aggregate, simply use the row index of the delta table as ordinal.
				ordinal = _rowIndex;
			}

			var dataPoint = dataPointCache && dataPointCache.get(ordinal);
			if (!dataPoint) {
				// Create a new DataPoint
				var dataPointOffset = rowData.length;
				for (var _dataItemIndex2 = 0; _dataItemIndex2 < dataItems.length; _dataItemIndex2++) {
					var _dataItem2 = dataItems[_dataItemIndex2];
					if (_dataItem2.type === 'cont') {
						rowData.push(new Value(_dataItem2));
					} else if (_dataItem2 === rowsDataItem) {
						// The tuple index is the source row index
						rowData.push(_baseRowIndex);
					} else {
						// Simply use the tuple index
						rowData.push(_row[_dataItem2.columnIndex]);
					}
				}
				dataPoint = new DataPoint(dataItems, rowData, dataPointOffset);
				dataPoints.push(dataPoint);
				// If there is no aggregation, there is no point in putting the DataPoint in a cache;
				// the ordinal will never be hit again.
				if (dataPointCache) {
					dataPointCache.set(ordinal, dataPoint);
				}
			} else if (!hasAggregatedData) {
				// Adding values to an existing data point: the data points contain aggregated data
				hasAggregatedData = true;
			}

			for (var _dataItemIndex3 = 0; _dataItemIndex3 < dataItems.length; ++_dataItemIndex3) {
				var _dataItem3 = dataItems[_dataItemIndex3];
				if (_dataItem3.type === 'cont') {
					dataPoint.getValue(_dataItemIndex3).add(_row[_dataItem3.columnIndex]);
				}
			}
		});

		for (var _dataItemIndex4 = 0; _dataItemIndex4 < dataItems.length; ++_dataItemIndex4) {
			var _dataItem4 = dataItems[_dataItemIndex4];
			if (_dataItem4.type === 'cont') {
				_dataItem4.setAggregationState(hasAggregatedData);
			}
		}

		return new VIPRDataSet(dataItems, dataPoints, _dataSetDef.slots, mapping);
	}

	// return the object that defines this module
	return {
		isInputColumn: isInputColumn,
		isTargetColumn: isTargetColumn,

		RowsColumn: RowsColumn,

		createTable: function createTable(_resultData, _slots) {
			var resultItemList = _resultData.getResultItemList();
			var columns = _resultData.getResultItemList().map(function (resultItem) {
				var mappingInfo = _slots.getMappingInfo(resultItem.getId());
				// Use the slot name as "kind". This will be 'context', 'input' or 'target'.
				var kind = mappingInfo.slot.getId();
				// No stacking support, so take the only column
				var sourceDataItem = resultItem.getDataItemList()[0];

				// possible values: 'integer' 'decimal', 'double', 'float', 'string', 'date', 'datetime', 'time'
				switch (sourceDataItem.getDataType()) {
					case 'integer':
					case 'decimal':
					case 'double':
					case 'float':
						return new MeasureColumn(sourceDataItem.getLabel(), sourceDataItem.getColumnId(), kind, sourceDataItem.getAggregation(), new SpecFormatter(sourceDataItem.getFormat()));

					// 'date', 'datetime', 'time' should not be features, but if they are, we consider them as categorical columns
					default:
						{
							// Odd API: the tuple count is reported via getRowCount().
							var tupleCount = resultItem.getRowCount();
							var tuples = new Array(tupleCount);
							for (var i = 0; i < tupleCount; i++) {
								var sourceTuple = resultItem.getValue(i);
								var sourceItem = sourceTuple[0];
								tuples[i] = new Tuple(sourceItem.d, sourceItem.u);
							}
							return new CategoricalColumn(sourceDataItem.getLabel(), sourceDataItem.getColumnId(), kind, Object.freeze(tuples));
						}
				}
			});

			// Process data into a new format: values that are either a tuple index or the actual value.
			var sourceDataPoints = _resultData.getDataPointValueList();
			// The data is stored in a flattened (1 dimensional) Array, which is a memory optimization.
			// Initializing this array with the expected length prevents the need to grow the Array's backing storage.
			var rowData = new Array(sourceDataPoints.length * columns.length);
			var insertRowIndex = 0;
			for (var rowIndex = 0; rowIndex < sourceDataPoints.length; rowIndex++) {
				// Track whether the row should be included.
				// Since AI models cannot handle missing values this is primarily based on whether the row has missing values.
				var includeRow = true;
				// Do not stop processing a row if includeRow turns to false; Tuples may still be discovered.
				for (var columnIndex = 0; columnIndex < columns.length; columnIndex++) {
					var column = columns[columnIndex];
					// Note that for every non aggregated column, value will be a tuple index.
					var value = sourceDataPoints[rowIndex].pt[columnIndex];
					if (column.type === 'measure') {
						// Reading the value gets a bit odd; due to the aggregate none measures are actually reported
						// as tuples, for any column but the target column.
						var specObj = resultItemList[columnIndex].getValue(value)[0];
						value = typeof specObj !== 'undefined' ? specObj.d : value.v;
						if (isMissingValue(value)) {
							value = null; // normalize missing values to null
							includeRow = false;
						}
					}
					// No else (column.type === 'categorical') required, use tuple index as value.
					rowData[insertRowIndex * columns.length + columnIndex] = value;
				}
				if (includeRow) {
					// only count the row if it is included.
					insertRowIndex++;
				}
			}
			// In case any rows where filtered out: trim the row data to the correct length.
			rowData.length = insertRowIndex * columns.length;

			return new SourceTable(columns, rowData);
		},

		/**
   * This method will take a source table, input deltas and a results Array, and create a virtual view that
   * combines the sources and the predicted rows.
   * @param {SourceTable} _sourceTable The SourceTable containing the base data.
   */
		createDeltaTable: function createDeltaTable(_sourceTable) {
			return new DeltaTable(_sourceTable.columns, _sourceTable.getRowData());
		},


		/**
   * Create a ColumnMapping object, that allows mapping columns to slots. This mapping can be used to run a
   * "query" on a table to build visualization data.
   * @param {VizDef} _vizDef The VizBundle vizDef.
   * @returns {ColumnMapping} The ColumnMapping that allows mapping columns to slots..
   */
		createColumnMapping: function createColumnMapping(_vizDef) {
			return new ColumnMapping(_vizDef);
		},


		/**
   * Create VIPR (visualization) data from a DeltaTable and slot mapping
   * @param {DeltaTable} _table The table to iterate.s
   * @param {ColumnMapping} _columnMapping The column mapping to apply.
   * Unmapped categoricals will be removed, aggregating values.
   */
		createViprData: function createViprData(_table, _columnMapping) {
			var dataSets = _columnMapping.vizDef.dataSets.map(function (_dataSetDef) {
				return queryDataSet(_dataSetDef, _table, _columnMapping);
			});
			return new VIPRData(dataSets);
		}
	};
});
//# sourceMappingURL=DataSet.js.map