'use strict';

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

/**
 * Licensed Materials - Property of IBM
 * IBM Cognos Products: BI Cloud (C) Copyright IBM Corp. 2020
 * US Government Users Restricted Rights - Use, duplication or disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 */
define(['@waca/dashboard-common/js/core/APIFactory', '@waca/dashboard-livewidget/js/features/content/dataQueryExecution/api/QueryModifierAPI'], function (APIFactory, QueryModifierAPI) {
	var QueryModifier = function () {
		function QueryModifier(_ref) {
			var content = _ref.content,
			    features = _ref.features;

			_classCallCheck(this, QueryModifier);

			this.content = content;
			features.DataQueryExecution.registerQueryModifier(this.getAPI());
		}

		QueryModifier.prototype.getAPI = function getAPI() {
			if (!this.api) {
				this.api = APIFactory.createAPI(this, [QueryModifierAPI]);
			}
			return this.api;
		};

		QueryModifier.prototype.getType = function getType() {
			return 'main';
		};

		QueryModifier.prototype.modifyQuerySpecList = function modifyQuerySpecList(querySpecList) {
			// Always Remove Aggregation 
			if (this._isTypeWhatIf()) {
				//if (this._isTypeWhatIf() && this._shouldRemoveAggregation()) {
				this._removeAggregation(querySpecList);
			}
			return querySpecList;
		};

		QueryModifier.prototype._isTypeWhatIf = function _isTypeWhatIf() {
			return this.content.getFeature('Visualization').getType() === 'WhatIf';
		};

		QueryModifier.prototype._shouldRemoveAggregation = function _shouldRemoveAggregation() {
			return !this._isContextSlotMapped() || this._isContextSlotMapped() && this.__hasStringDataItemInInputSlot();
		};

		QueryModifier.prototype._removeAggregation = function _removeAggregation(querySpecList) {
			querySpecList.forEach(function (querySpec) {
				querySpec.spec.dataItems.forEach(function (dataItem) {
					if (dataItem.aggregate) {
						dataItem.aggregate = 'none';
					}
				});
			});
		};

		QueryModifier.prototype._isContextSlotMapped = function _isContextSlotMapped() {
			var result = false;
			var mappedSlots = this.content.getFeature('Visualization').getSlots().getMappedSlotList();
			mappedSlots.forEach(function (slot) {
				if (slot.getId() === 'context') {
					result = true;
				}
			});
			return result;
		};

		QueryModifier.prototype.__hasStringDataItemInInputSlot = function __hasStringDataItemInInputSlot() {
			var result = false;
			var mappedSlots = this.content.getFeature('Visualization').getSlots().getMappedSlotList();
			var index = mappedSlots.findIndex(function (slot) {
				return slot.getId() === 'input';
			});
			var inputSlot = mappedSlots[index];
			inputSlot.getDataItemList().forEach(function (dataItem) {
				if (dataItem.getDataType() === 'string') {
					result = true;
				}
			});
			return result;
		};

		return QueryModifier;
	}();

	return QueryModifier;
});
//# sourceMappingURL=WhatIfQueryModifier.js.map